using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Text;			// for Encoding I/O to a Text file
using System.Windows.Forms;	// for RichTextBox
using System.Threading;		// for Thread

namespace MyNamespace
{
	/// <summary>
	///		Common part of CrypF, a data encryption system, Forms version
	///		from Common routines for calculator programs
	///		Developed in TURBO Pascal 5.5 and 6.0, converted to Visual Basic 6
	///		and now in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	public class Common
	{
		public Common()		// constructor logic
		{
		}

		public static void RunCalc(string st)	// run on a Text stream inS
		{
			StreamReader inCF = null;			// run from inS (input screen) vice inCF input (code file)
			RunCalc(inCF, st);
		}

		// ------------------ the main command processor
		private static void RunCalc(StreamReader inCF, string st) // run on a text stream or a code file
		// byFile is true if input is from File, not by a TextBox
		{
			int i;
			int resSav;
			string ch;
			string st8;
			string fName;
			bool done = false;
			MultiID.muAbort = false;
			nested++;
			bool byFile = (inCF != null);
			if (!byFile) nested = 0;
			do
			{
				noChange = false;
				MultiID.muErr = false;
				if (byFile)
				{
					ReadCom(inCF, out st);	// read next command from file
					if (st == null) break;	// null => EOF (end of file)
					if (st != "" && st[0] == '@') BatchOp(ref st);
					if (echoB)
					{
						MultiID.WriteTLn(Prompt() + " " + st); // echo to output textbox
						DoubleSpace();
					}
				}
				else
				{
					if (st == "")
						break;
					else
					{
						MultiID.WriteTLn(Prompt() + " " + st); // echo command as a prompt line
						DoubleSpace();
					}
				}
				GC.Collect();
				if ((echoB || !byFile) && MultiID.timeIt)
				{
					long totalMemory = GC.GetTotalMemory(false);
					DoubleSpace();
					MultiID.TimeIt("Start of execution, Memory in use = " + totalMemory);
				}
				st8 = st.Substring(0, Math.Min(8, st.Length));
				st8 = st8.ToUpper();
				prSt = st.Trim(' ');	// needed for GoUpTo Label
				if (!byFile)
				{
					i = 0;				// delete line continuation markers in TextBox input
					do
					{
						i = st.IndexOf("\n", i + 1);
						if (i < 0) break;
						ch = st.Substring(i - 1, 1);
						if (ch == "+")
						{
							ReplaceAt(ref st, i - 1, "  "); // +\n -> 2 spaces
						}
						else if (ch == "-")
						{
							MultiID.Delete(ref st, i - 1, 2);
							i -= 2;		// -CrLf deleted and string closed up
						}
						else
						{
							ReplaceAt(ref st, i, ' '); // \n -> 1 space
						}
					}
					while (true);
				}
				st = st.Trim(' ');	// one final trim
				while (st != "")	// process next part on command line
				{
					char ch1 = st[0];
					string op = "";
					if (st.Length > 1) op = st.Substring(0, 2);
					resSav = 0;		// to prevent unassigned local variable alarm
					fName = "";		// same thing
					if (op == "//") st = "";
					else if (BatchOps(ref st)) {}									// a Batch command was executed
					else if (SkipLabel(ref st)) {}								// a Label was skipped
					else if (GoToLabel(ref prSt, ref st, inCF)) {}				// goTo statement was processed
					else if (Calc.AProcedure(ref st, ref resSav, ref fName))	// a procedure was processed
						{Calc.ChkResSav(resSav, fName);}
					else if (Calc.Assignment(ref st)) {}						// assignment was made
					else if (Calc.Primitive(ref st, 0)) { }						// a Primitive was found
					else
					{
						MultiID.WriteTDLn("Unknown operation, Command line discarded: |" + st);
						st = "";
					}
					MultiID.MuInterrupt();
					if (MultiID.muAbort)
					{
						st = "";
					}
					st = st.TrimStart(' ', ';');	// delete leading and trailing delimiters
					GC.Collect();	// force garbage collection of all generations
				} // end of while (st != "")
				if (echoB || !byFile)
				{
					long totalMemory = GC.GetTotalMemory(false);
					MultiID.TimeIt("End of execution, Memory in use = " + totalMemory);
				}
				if (byFile) done = (st == null || MultiID.muAbort);
				MultiID.muErr = false;
				if ((echoB || !byFile))
				{
					DoubleSpace();
				}
			}
			while (!done && !MultiID.muAbort);
			nested--;
			// trap for error handling is in calling routine
		} // RunCalc

		public static void SingleSpace() // guarantee a single line space in outS
		{
			SingleSpace(MultiID.outS);
		}

		public static void SingleSpace(RichTextBox outB) // guarantee a single line space in output box
		{
			int iL = outB.Text.Length;
			if (iL > 0)
			{
				if (outB.Text.Substring(iL - 1) != "\n") MultiID.WriteLn(outB, null);
			}
		}

		public static void DoubleSpace(RichTextBox outB) // guarantee a double line space in output box
		{
			int iL = outB.Text.Length;
			if (iL > 0)
			{
				if (outB.Text.Substring(iL - 1) != "\n") MultiID.WriteLn(outB, null);
				if (outB.Text.Substring(outB.Text.Length - 2) != "\n\n") MultiID.WriteLn(outB, null);
			}
		}

		public static void DoubleSpace() // guarantee a double line space in outS
		{
			DoubleSpace(MultiID.outS);
		}

		private static string Prompt() // construct Command prompt
		{
			// comHL(1000000) = 1 // diag
			string st;
			DoubleSpace();
			st = ":";
			if (nested == 0)
			{
				st = "Command" + st;
			}
			else
			{
				st = "Command" + nested.ToString() + st;
			}
			return st;
		}

		public static bool DeleteComment(ref string st, out string stCom)
		// delete comment, Rem, Pause, and Echo
		// if nothing found, st not changed and stCom = ""
		// if found, st reduced and stCom = comment if one found
		{
			int i;
			string ch;
			string field;
			stCom = "";
			if (st[0] == '\"')
			{
				i = st.IndexOf("\"", 1) + 1;
				if (i == 0) i = st.Length;
				stCom = st.Substring(0, i);
				DelNPBls(ref st, i);
				return true;
			}
			field = GetField(ref st);
			i = field.Length + 1;
			if (st.Length < i)
			{
				ch = "=";
			}
			else
			{
				ch = st.Substring(i - 1, 1);
			}
			if (field == "ECHO" || field == "PAUSE" || field == "REM")
			{
				st = "";
				noChange = false; // not in DOS VPCalc
			}
			else
			{
				return false;
			}
			return true;
		} // DeleteComment

		public static bool BatchOp(ref string st)	// process Batch operator @ECHO OFF
		{
			string field2;
			string field1 = GetField(ref st);		// @ is a delimiter so do special processing
			if (field1 != "") return false;
			int iL = st.Length;
			field1 = st.Substring(0, Math.Min(6, iL));
			field1 = field1.ToUpper();
			if (iL > 5 && inDelim(field1[5])) ReplaceAt(ref field1, 5, ' ');
			if (field1 != "@ECHO " && field1 != "@ECHO") return false;
			DelNPBls(ref st, 5);
			field2 = GetField(ref st);				// ECHO "" is comment
			if (field2 == "OFF" && st.Length == 3)
			{
				echoB = false;
			}
			else if (field2 == "ON" && st.Length == 2)
			{
				echoB = true;
			}
			st = "";
			return true;
		} // BatchOp

		public static bool BatchOps(ref string st)
		// process Batch operators ECHO, REM and PAUSE
		{
			string field1;
			string field2;
			if (BatchOp(ref st))
			{
				return true;
			}
			field1 = GetField(ref st);
			if (field1 == "ECHO")
			{
				DelNPBls(ref st, 4);
				field2 = GetField(ref st);	// ECHO "" is comment
				if (field2 == "OFF" && st.Length == 3)
				{
					echoB = false;
				}
				else if (field2 == "ON" && st.Length == 2)
				{
					echoB = true;
				}
				else if (echoB)
				{
					MultiID.WriteTLn2(st);
				}
			}
			else if (field1 == "PAUSE")
			{
				Global.objRunForm.WaitForPause();
			}
			else if (field1 != "REM")
			{
				return false;	// not ECHO, PAUSE or REM
			}
			st = "";
			return true;
		} // BatchOps

		public static bool SkipLabel(ref string st)	// skip a Label
		{
			string label1 = GetField(ref st);
			int i = label1.Length;
			if (i < st.Length && st[i] == ':')
			{
				DelNPBls(ref st, i + 1);
				MultiID.ADiag("Label skipped = " + label1 + ":");
				return true;
			}
			return false;
		}

		public static bool GoToLabel(ref string prSt, ref string st, StreamReader inCF)	// do a GoTo statement
		{
			bool upTo;
			string label1;
			bool findLabel;
			bool byFile = (inCF != null);
			string st7 = GetField(ref st);
			if (st7 != "GOTO" && st7 != "GOUPTO") return false;
			upTo = (st7 != "GOTO");
			if (upTo)
			{
				DelNPBls(ref st, 6);
			}
			else
			{
				DelNPBls(ref st, 4);
			}
			label1 = GetField(ref st);
			if (upTo) st = prSt;
			LookLabel(ref st, label1, out findLabel);
			if (findLabel) return true;
			if (!byFile) return true;
			if (upTo)
			{
				inCF.BaseStream.Seek(0, SeekOrigin.Begin);	// rewind inCF to beginning of the file
				inCF.DiscardBufferedData(); // call inCF.DiscardBufferedData() after you perform the Seek() operation
											// so the stream reader does not get stale data from its buffer.
											// This from mikeb in news group microsoft.public.dotnet.languages.csharp 5/14/2004
			}
			do
			{
				st = inCF.ReadLine(); // read in next line
				if (st == null) { st = ""; break; } // unexpected end of file, label not found
				DelTabs(ref st);
				LookLabel(ref st, label1, out findLabel);
				if (findLabel) return true;
				MultiID.MuInterrupt();
			}
			while (!MultiID.muAbort);
			return true;
		} // GoToLabel

		public static void LookLabel(ref string st, string label1, out bool findLabel)
		{
			string label2;
			findLabel = false;
			while (st != "")
			{
				if (!DeleteComment(ref st, out label2))
				{
					label2 = GetField(ref st);			// comment was not deleted
					int i = label2.Length;
					if (st.Length > i && st[i] == ':')	// test Label
					{
						DelNPBls(ref st, i + 1);
						if (label1 == label2)
						{
							if (echoB)
							{
								MultiID.WriteTLn2("Went to " + label1 + ": " + st);
							}
						findLabel = true;
						return;
						}
					}
					else
					{
						DelNPBls(ref st, i + 1);
					}
				}
			}
		}

		public static bool OpenRead(out StreamReader inF, string fName, out string dir)
		// open file for Reading from CurrentDirectory or homeDir
		{
			string stEx1;
			string stEx2 = "";
			string savDir = Environment.CurrentDirectory.ToString();
			if (!quiet) DoubleSpace();
			dir = savDir;
			inF = null;
			if (OpenReadFile(out inF, fName, out stEx1))
			{
				if (!quiet)
				{
					MultiID.WriteLn("File \"" + fName + "\" opened for reading");
					WCurDir(fName);
				}
				return true;
			}
			if (savDir != homeDir)
			{
				dir = homeDir;
				Directory.SetCurrentDirectory(homeDir);
				if (OpenReadFile(out inF, fName, out stEx2))
				{
					if (!quiet)
					{
						MultiID.WriteLn("File \"" + fName + "\" opened for reading");
						WCurDir(fName);
					}
					Directory.SetCurrentDirectory(savDir);
					return true;
				}
				Directory.SetCurrentDirectory(savDir);
			}
			dir = "";
			MultiID.WriteLn("Cannot open file \"" + fName + "\" for input");
			MultiID.WriteTLn("Not in current directory " + savDir);
			ReportError("OpenRead", stEx1);		// file not found most likely
			if (savDir != homeDir)
			{
				MultiID.WriteTLn("Not in home directory " + homeDir);
				ReportError("OpenRead", stEx2); // file not found most likely
			}
			return false;
		} // OpenRead

		public static bool OpenReadFile(out StreamReader inF, string fName, out String stEx)
		// open file for Reading from CurrentDirectory
		{
			inF = null;
			stEx = "";
			try
			{
				inF = new StreamReader(fName); // open fName for input as inF
				return true;
			}
			catch (Exception ex)
			{
				stEx = ex.Message.ToString();
				return false;
			}
		}

		public static bool OpenWrite(out StreamWriter outF, string fName) // open file for Writing
		{
			outF = null; // outF must be assigned to before control leaves the current method
			try
			{
				// open fName for output as outF
				outF = new StreamWriter(fName, false, Encoding.ASCII); // Appending = false
				if (!quiet)
				{
					MultiID.WriteDLn("File \"" + fName + "\" opened for writing");
					WCurDir(fName);
				}
				return true;
			}
			catch (Exception ex)
			{
				MultiID.WriteDLn("Cannot open file \"" + fName + "\" for output");
				WCurDir(fName);
				ReportError("OpenWrite", ex.Message);
				return false;
			}
		}

		public static bool ChkWrite(out StreamWriter outF, string fName, string st)
		// check start write of file
		{
			if (!OpenWrite(out outF, fName))
			{
				if (MakeReady() == DialogResult.OK)
				{
					if (!OpenWrite(out outF, fName))
					{
						MultiID.WriteLn("Aborting file write...");
						return false;
					}
				}
			}
			if (!quiet) MultiID.WriteTLn("Writing file \"" + fName + "\"" + st);
			return true;
		}

		public static bool OpenAppend(ref StreamWriter outF, string fName)	// open file for Writing
		{
			try
			{
				// open fName for Append as outF
				outF = new StreamWriter(fName, true, Encoding.ASCII);		//Appending = true
				return true;
			}
			catch (Exception ex)
			{
				MultiID.WriteDLn("Cannot open file \"" + fName + "\" for output");
				WCurDir(fName);
				ReportError("OpenAppend", ex.Message);
				return false;
			}
		}

		public static bool ChkAppend(ref StreamWriter outF, string fName)
		// check start Append to a file
		{
			if (!OpenAppend(ref outF, fName))
			{
				if (MakeReady() == DialogResult.OK)
				{
					if (!OpenAppend(ref outF, fName))
					{
						MultiID.WriteLn("Aborting file write...");
						return false;
					}
				}
			}
			return true;
		}

		public static void ReportError(string who, string errorMessage) // report IOResult etc. error, errorMessage = caught.Message
		{
			GC.Collect();	// force garbage collection of all generations
			MultiID.muErr = true;
			if (errorMessage == "") return;
			if (MultiID.muErrRep) MultiID.WriteTLn(who + ": " + errorMessage);
		}

		public static DialogResult MakeReady() // pause for operator to make file/drive ready
		{
			DialogResult key = MessageBox.Show(Global.pName + ": Make file/drive ready and click \"OK\"\n" +
								StringOf(29, " ") + "or \"Cancel\"",
				Global.title,
				MessageBoxButtons.OKCancel,
				MessageBoxIcon.Exclamation);
			// pause ----------
			if (key != DialogResult.OK) MultiID.WriteLn("I/O operation aborted by operator!");
			return key;
		}

		public static void ChangeDir(ref string st)
		// change the default disk directory and remove its name from command line
		{
			string stL;
			string savDir;
			if (st.Length == 0) st = "\""; // for safety
			stL = FileNm(ref st);
			savDir = "";
			try
			{
				if (!quiet) DoubleSpace();
				savDir = Environment.CurrentDirectory.ToString();
				if (stL == savDir) stL = "";
				if (stL != "") Directory.SetCurrentDirectory(stL);
			}
			catch
			{
				// -------- Trap for error handling
				MultiID.WriteLn("Path not found!");
			}
			finally
			{
				if (!quiet)
				{
					if (Environment.CurrentDirectory.ToString() != savDir)
					{
						MultiID.WriteTLn("Directory changed to \"" + Environment.CurrentDirectory.ToString() + '"');
					}
					else
					{
						MultiID.WriteTLn("Directory not changed \"" + Environment.CurrentDirectory.ToString() + '"');
					}
					MultiID.WriteTLn("Home directory is \"" + homeDir + '"');
				}
			}
		}

		public static string FileNm(ref string st)
		// return and remove file name from command line
		{
			int i;
			string stL;
			if (st.Length == 0)
			{
				return "";
			}
			i = st.IndexOf("\"", 1);
			if (i <= 0) i = st.Length;
			stL = st.Substring(1, i - 1);
			DelNPBls(ref st, i); // do not delete the trailing " if any
			return stL.Trim();
		}

		public static void ToggleLogg() //Toggle Logging of output screen to Log File
		{
			DateTimePicker dateTime = new System.Windows.Forms.DateTimePicker();
			string dateTimeString = dateTime.Value.ToString();
			if (!quiet) DoubleSpace();
			if (MultiID.logg)
			{
				string st = "Log file \"" + logFile + "\" Closed " + dateTimeString;
				if (!quiet)
				{
					MultiID.WriteLn(st);
					MultiID.outL.WriteLine();
				}
				MultiID.outL.Close();	// close the file
				logFile = "";			// allow for getting a new log file name next open
				MultiID.logg = false;
			}
			else
			{
				Calc.CalLFile(ref logFile); // get log file name
				bool check;
				check = ChkAppend(ref MultiID.outL, logFile);
				if (check) // open logFile for Append as MultiID.outL
				{
					MultiID.logg = true;
					string st = "Log file \"" + logFile + "\" Opened for Append " + dateTimeString;
					if (!quiet)
					{
						MultiID.WriteDLn(st);
						WCurDir(logFile);
						MultiID.outL.WriteLine(Global.title + ", " + Global.version);
						if (Global.system != "") MultiID.outL.WriteLine("Run on: " + Global.system);
					}
				}
				else
				{
					logFile = ""; // file did not open
					return;
				}
			}
			Global.objRunForm.SetLogCaption(); // call sub in Run form
		} // ToggleLogg

		public static void DelTabs(ref string st)
		// delete leading and trailing blanks, change Tabs to spaces, for code files
		// and DOS command line parameters
		// side effect, set noChange false to init processing on command
		{
			st = st.Replace("\t", " ");			// all Tab -> Space
			int i = -1;							// delete Tabs in Calc code files
			do
			{
				i = st.IndexOf('\t', i+1);		// index of a horizontal tab character
				if (i >= 0)
				{
					ReplaceAt(ref st, i, ' ');	// Tab -> Space
				}
			}
			while (i >= 0);
			st = st.Trim();						// trim all white space from start and end of command
			noChange = false;
		}

		public static void ReSizeHist(int newMax)	// change length of command history list
		{
			if (newMax == mComs || newMax < 1)
			{
				if (!quiet) MultiID.WriteDLn("Max commands in history not changed from " + mComs);
				return;
			}
			int d = cComs - newMax;					// d = number of commands to delete
			if (d < 0) d = 0;
			cComs = Math.Min(cComs, newMax);		// number of commands left
			string[] tComHL = new string[cComs+1];	// re-dimention arrays to 1 to newMax
			string[] tComPF = new string[cComs+1];
			iComs = cComs; // command to restore
			for (int j = 1; j <= cComs; j++)		// save commands in a temp array
			{
				tComHL[j] = comHL[j + d];			// delete oldest commands if d > 0
				tComPF[j] = comPF[j + d];
			}
			comHL = new string[newMax+1];			// re-dimention arrays to 1 to newMax
			comPF = new string[newMax+1];
			for (int j = 1; j <= cComs; j++)
			{
				comHL[j] = tComHL[j];				// copy cComs commands
				comPF[j] = tComPF[j];
			}
			if (!quiet) MultiID.WriteDLn("Max commands in history changed from " + mComs + " to " + newMax);
			mComs = newMax;
		}

		public static void ClearHist()		// clear command history list
		{
			cComs = 0;
			comHL = new string[mComs+1];	// re-dimention arrays
			comPF = new string[mComs+1];
			if (!quiet)
			{
				MultiID.WriteDLn("Command history cleared");
				RunForm.FocusBothUs();
			}
		}

		public static void ClearLog()	// clear Log file
		{
			DateTimePicker dateTime = new System.Windows.Forms.DateTimePicker();
			string dateTimeString = dateTime.Value.ToString();
			bool logS = MultiID.logg;	// save Log status in a local
			if (!quiet) DoubleSpace();
			if (MultiID.logg)
			{
				string st = "Log file \"" + logFile + "\" Closed " + dateTimeString;
				if (!quiet)
				{
					MultiID.WriteLn(st);
					MultiID.outL.WriteLine();
				}
				MultiID.WriteLn(null, MultiID.outL);
				MultiID.outL.Close();		// close the file
				MultiID.logg = false;
			}
			logFile = "";
			Calc.CalLFile(ref logFile);	// get log file name, it may have changed
			// open logFile for output as outL, clear the file, outL gotten by open routine
			if (ChkWrite(out MultiID.outL, logFile, ""))
			{
				MultiID.logg = true;		// set to get message to the file
				string st = "Log file \"" + logFile + "\" Cleared " + dateTimeString;
				if (!quiet)
				{
					MultiID.WriteLn(st);
					WCurDir(logFile);
				}
				else
				{
					MultiID.outL.WriteLine(st);
				}
				MultiID.outL.WriteLine(Global.title + ", " + Global.version);
				if (Global.system != "") MultiID.outL.WriteLine("Run on: " + Global.system);
				if (!quiet)
				{
					MultiID.WriteLn();
				}
				else
				{
					MultiID.outL.WriteLine();
				}
				if (!logS)
				{
					MultiID.outL.Close();	// close the file
					MultiID.logg = false;
					logFile = "";			// clear name when it is closed, not if its open
				}
			}
		} // ClearLog

		public static void AddCom(string st, string sP)
		// add a command to the command history list
		{
			int j;
			bool found;		// true if command found in history list
			if (st == "") return;
			found = false;
			for (int i = cComs; i >= 1; i--)		// check newest first
			{
				if (st == comHL[i])					// if command already in history list
				{
					found = true;
					sP = comPF[i];					// save and use protect flag from old command
					for (j = i; j < cComs; j++)		// delete it from list
					{
						comHL[j] = comHL[j+1];		// freeing up comHL[cComs]
						comPF[j] = comPF[j+1];
					}
					break;
				}
			}
			if (!found)
			{
				cComs = cComs + 1;
				if (cComs > mComs)					// if too many commands, delete oldest command
				{
					cComs = mComs;
					for (j = 1; j < cComs; j++)
					{
						comHL[j] = comHL[j+1];		// free up comHL(cComs)
						comPF[j] = comPF[j+1];
					}
				}
			}
			comHL[cComs] = st;						// add command to top of list
			comPF[cComs] = sP;
			iComs = cComs;
		} // AddCom

		public static void RestoreHist() // read history file into command list
		{
			string st;
			string st2;
			string sP;
			int i;
			StreamReader inF = null;
			string dir = null;
			int NumRead = 0;
			if (!quiet) DoubleSpace();
			if (!OpenRead(out inF, histFile, out dir)) return;
			if (!quiet) MultiID.Write("Reading Commands from History file - Begin ... ");
			try
			{
				st = "";
				while (st != null && st.IndexOf("1: ") == -1) // test first to allow for an empty file
				{
					st = inF.ReadLine();		// get first string of first command
				}
				while (st != null && cComs <= comHL.GetUpperBound(0)) // loop until end of file or array is full
				{
					sP = st.Substring(0, 1);	// get protect flag
					st2 = "";
					i = st.IndexOf(" ", 1);
					st = st.Substring(i + 1);	// delete through first blank
					bool done = false;
					do
					{
						if (st2 != "") st2 += "\n";
						st2 += st;			// append next string to command
						st = inF.ReadLine();	// get the next string in command
						if (st != null && st.Length > 3 && st.Substring(0, 4) == "----")
							done = true;
					}
					while (st != null && !done); // until end of a command
					if (st != null)
					{
						AddCom(st2, sP);		// use this protect flag if not a duplicate command
						NumRead++;
						st = "";
						st2 = "";
						st = inF.ReadLine();	// get first string of next command
					}
				}
				if (!quiet)
				{
					MultiID.WriteLn("End");
					MultiID.Write(NumRead + " Command");
					if (NumRead != 1) MultiID.Write("s");
					MultiID.WriteLn(" read from History file, " + cComs + " total");
				}
			}
			catch (Exception ex)
			{
				ReportError("RestoreHist", ex.Message);
				MultiID.WriteLn("Aborted");
				MultiID.WriteLn("File is corrupted: \"" + histFile + '"');
			}
			finally
			{
				if (inF != null) // if the file was opened
				{
					inF.Close(); // close the file
					RunForm.FocusBothUs();
				}
			}
		} // RestoreHist

		public static void SaveHist() // save command history to file X?CalcHist.txt
		{
			int i;
			StreamWriter outF = null;
			string savDir;
			DoubleSpace();
			savDir = Environment.CurrentDirectory.ToString();
			Directory.SetCurrentDirectory(homeDir);
			DateTimePicker dateTime = new System.Windows.Forms.DateTimePicker();
			string dateTimeString = dateTime.Value.ToString();
			try
			{
				FileInfo FIa = new FileInfo(histFile);
				FileInfo FIb = new FileInfo(histBak);
				if (FIa.Exists)
				{
					FIb.Delete();			// delete the backup if any
					FIa.MoveTo(histBak);	// save (rename) previous History file as a backup
				}
			}
			catch (Exception ex)
			{
				ReportError("SaveHist", ex.Message);
			}
			try
			{
				outF = new StreamWriter(histFile);	// open the history file
				if (!quiet)
				{
					MultiID.WriteDLn("File " + histFile + " opened for writing");
					WCurDir(histFile);
					MultiID.Write("Writing all Commands to History file - Begin ... ");
				}
				outF.WriteLine("History of Command Input - " + dateTimeString);
				outF.WriteLine(Global.title + ", " + Global.version);
				if (Global.system != "") outF.WriteLine("Run on: " + Global.system);
				outF.WriteLine("");
				for (i = 1; i <= cComs; i++)		// write all previous commands
				{
					outF.WriteLine(comPF[i] + i + ": " + comHL[i]);
					outF.WriteLine(StringOf(40, "-"));
				}
				outF.WriteLine("End of file " + histFile);
				if (!quiet)
				{
					MultiID.WriteLn("End");
					MultiID.Write(cComs + " Command");
					if (cComs != 1) MultiID.Write("s");
					MultiID.WriteLn(" written to History file");
				}
				Directory.SetCurrentDirectory(savDir);
			}
			catch (Exception ex)
			{
				ReportError("SaveHist", ex.Message);
			}
			finally
			{
				if (outF != null)
				{
					outF.Close();					// close the file
					RunForm.FocusBothUs();
				}
			}
		} // SaveHist

		public static void FixFName(ref string fName)
		// fix file name to path + name
		{
			int i;
			if (fName == null)
				fName = "";
			if (fName != "" && fName[0] == '"')
			{
				i = fName.IndexOf('"', 1);
				if (i >= 0) fName = fName.Substring(0, i);	// deleted all but "..."
				fName = fName.Trim('"');
			}
			fName = fName.Trim(' ', '.', '\\');				// remove leading and trailing blanks '.' and '/'
			if (fName == "") return;
			int p = fName.IndexOf(@"\");				// find last "\"
			i = fName.IndexOf('.', p + 1);				// a "." after the last "\" starts extension
			i = fName.IndexOf(":.");					// :. OR \. is a bad file name
			if (i < 0) i = fName.IndexOf(@"\.");
			if (i >= 0) fName = fName.Substring(0, i);	// if \.xxx, or :.xxx, .xxx deleted
		}

		public static void FixFName(ref string fName, string ext)
		// fix file name to path + name + .extension
		{
			if (ext == "")
			{
				FixFName(ref fName);
				return;
			}
			int i;
			if (fName == null)
				fName = "";
			if (fName != "" && fName[0] == '"')
			{
				i = fName.IndexOf('"', 1);
				if (i >= 0) fName = fName.Substring(0, i);	// deleted all but "..."
				fName = fName.Trim('"');
			}
			fName = fName.TrimStart(' ', '.', '\\');			// remove leading banks '.' and '\'
			fName = fName.TrimEnd('\\');						// remove trailing '\'
			int j = -1;
			while (true)
			{
				i = fName.LastIndexOf('.');
				if (i == j || i < 0)
					break;
				j = i;
				fName = fName.TrimEnd(' ', '\\');				// remove blanks and '\' after '.'
				fName = fName.TrimEnd('.', '\\');				// remove trailing '.' and '\'
			}
			if (fName == "") return;
			int p = fName.IndexOf(@"\");				// find last "\"
			i = fName.IndexOf('.', p + 1);				// a "." after the last "\" starts extension
			if (i < 0)
				fName = fName + "." + ext;				// use default extension
			i = fName.IndexOf(":.");					// :. OR \. is a bad file name
			if (i < 0) i = fName.IndexOf(@"\.");
			if (i >= 0) fName = fName.Substring(0, i);	// if \.xxx, or :.xxx, .xxx deleted
		}

		public static void WCurDir(string fName)
		// write current disk directory with final \
		// using Environment.CurrentDirectory.ToString(): C:\ already has the \ but non-root directories do not
		{
			// this can garble some output, but good info
			string cD = Directory.GetCurrentDirectory();
			if (cD.Substring(cD.Length - 1) != @"\") cD += @"\";
			if (fName.Length > 1 && fName[1] == ':')
			{
				MultiID.WriteTLn("Directory name = \"" + cD + '"');
			}
			else
			{
				MultiID.WriteTLn("Full name = \"" + cD + fName + '"');
			}
		}

		public static void ReadCom(StreamReader inCF, out string st)
		// read in one command line from open file inCF including continuation lines
		{
			string stL;
			string st1;
			char ch;
			bool bContinue;
			string cat = " ";
			st = ""; // for safety
			try
			{
				do
				{
					bContinue = false;
					stL = inCF.ReadLine();	// get line of command
					if (stL == null)
					{
						st = null;			// null => end of file
						return;
					}
					DelTabs(ref stL);
					if (stL.Length > 0)
					{
						ch = stL[stL.Length-1];
					}
					else
					{
						ch = ' ';
					}
					if (ch == '+' || ch == '-')
					{
						bContinue = true;
						ReplaceAt(ref stL, stL.Length - 1, ' ');
						stL = stL.TrimEnd(' ');
					}
					st1 = " ";
					if (stL == "")
					{
						st1 = "";
					}
					else if (inDelim(stL[0]))
					{
						st1 = "";
					}
					else if (st == "")
					{
						st1 = "";
					}
					else if (inDelim(st[st.Length-1]))
					{
						st1 = "";
					}
					if (st1 == " ") st1 = cat;
					if (ch == '+')
					{
						cat = " ";
					}
					else
					{
						cat = "";
					}
					if (st == "")
					{
						st = stL;
					}
					else
					{
						st = st + st1 + stL;
					}
				}
				while (bContinue);
			}
			catch
			{
				// -------- Trap for error handling
				st = null; // read error same as EOF
			}
		} // ReadCom

		public static bool inDelim(char ch)
		// is the character ch in the list of Delimiters
		{
			return (delim.IndexOf(ch) >= 0);
		}

		public static void DelNPBls(ref string st, int n)
		// delete 1st n characters plus new leading blanks, Side effect: noChange = false;
		{
			noChange = false;
			if (n >= st.Length) { st = ""; return; }
			st = st.Substring(n).TrimStart(' ');
		}

		public static void ReportSin(string message, ref string st)
		// report first syntax error, clear st
		{
			if (!MultiID.muErr)
				MultiID.WriteTDLn(message + st);
			MultiID.muErr = true;
			st = "";
		}

		public static string GetField(ref string st) // get and remove next field
		// assumes no leading blanks, no trailing blank, no tabs. Returns first string
		// in caps up to first delimiter or blank, may be null, deletes extra blanks
		// after first field in st, returned field not removed or changed.
		// noChange true until DelNPBls is called to delete the field from st
		{
			int i;
			int j;
			int iL;
			if (noChange) return oldField;
			noChange = true;
			iL = st.Length;
			if (iL == 0)
			{
				oldField = "";
				return oldField;
			}
			i = 0;
			while (i < iL && st[i] != ' ' && !inDelim(st[i])) i++;
			field = st.Substring(0, i);
			oldField = field.ToUpper();
			if (i >= iL || st[i] != ' ') return oldField;
			j = i + 1;
			while (j < iL && st[j] == ' ') j++;		// st: "" => "", "1" => "1"
			if (j < iL && !inDelim(st[j])) j--;		// "1	  2" => "1 2"
			MultiID.Delete(ref st, i, j-i);			// "1	;  2" => "1;  2"
			return oldField;
		}

		public static bool RunFile(string fName)
		{
			StreamReader r; // text file to read from
			string st;		// operator input string
			string dir;
			if (fName.Length == 0) return false; // noOp, would have input from TextBox
			if (!OpenRead(out r, fName, out dir)) return false;
			if (!quiet) MultiID.WriteLn("Running code file \"" + fName + '"');
			Global.objRunForm.AddToTitle(" " + '"' + fName + '"');
			st = "";
			RunCalc(r, st);
			r.Close();
			noChange = false;
			if (!quiet) MultiID.WriteDLn("File \"" + fName + "\" closed");
			Global.objRunForm.DelFromTitle(" " + '"' + fName + '"');
			return true;
		}

		public static void ReplaceAt(ref string oldS, int i, string newS) // change oldS at i to newS
		{
			oldS = oldS.Substring(0, i) + newS + oldS.Substring(i + newS.Length);
		}

		public static void ReplaceAt(ref string oldS, int i, char ch) // change oldS at i to ch
		{
			oldS = oldS.Substring(0, i) + ch + oldS.Substring(i + 1);
		}

		public static string StringOf(int n, string stIn) // repeat stIn, n times
		{
			string stOut = "";
			for (int i = 0; i < n; i++)
			{
				stOut += stIn;
			}
			return stOut;
		}

		public static void DisplayPriority() // display priority status
		{
			string st = Thread.CurrentThread.Priority.ToString();
			MultiID.WriteDLn("Priority is " + st);
		}

		public static void SetPriority(int p) // change priority to p
		{
			if (p == 2)
				Thread.CurrentThread.Priority = ThreadPriority.Highest;
			else if (p == 1)
				Thread.CurrentThread.Priority = ThreadPriority.AboveNormal;
			else if (p == -1)
				Thread.CurrentThread.Priority = ThreadPriority.BelowNormal;
			else if (p == -2)
				Thread.CurrentThread.Priority = ThreadPriority.Lowest;
			else
			{
				p = 0;
				Thread.CurrentThread.Priority = ThreadPriority.Normal;
			}
			pri = p;
		}

		// list of command history of operator input in a dynamic array
		public static string[] comHL;	// the saved command history list (1 to mComs)
		public static string[] comPF;	// the saved command history protect flag
		public static int mComs;		// max # of commands you can save, array size
		public static int cComs;		// current # of commands saved in history list
		public static int iComs;		// number of Command to restore to Text Input Box
		// end of Command History List

		public static bool aboutUp;		// true if About form is up and running
		public static string comI;      // comment item, its name has the comment/File name
		public static bool configUp;	// true if Config form is up and running
		public static string delim;		// delimiters, ;, +, ...
		public static bool echoB;		// true if Batch file echo flag is on
		public static string histBak;	// file name of command history backup file
		public static string histFile;	// file name of command history file
		public static bool historyUp;	// true if History form is up and running
		public static string homeDir;	// initial disk directory when program started
		public static RichTextBox inS;	// input Screen source of commands
		public static string logFile;	// file name of log file or "" if not logging
		public static int nested;		// number of nettings of the RunCalc procedure, 0 = bad
		public static bool noChange;	// true after GetField until DelNPBls is called
		public static string field;		// field gotten by GetField
		public static string oldField;	// last field gotten by GetField in uppercase, if noChange true
		public static int pri;			// execution priority [-2, 2]
		public static string prSt;		// previous operator input string, needed for GoUpTo Label
		public static bool quiet;		// true if "quiet" mode is on
		public static Button runCmdLog; // command button on Run form
		public static bool running;		// true if Calc is currently processing a command
		public static string selectPath;// path selected on the Config Form
		public static string selectFile;// file selected on the Config Form
		public static int selectFilter;	// file selected FilterIndex the Config Form
		public static bool tQuitReq;	// request to totally quit the program
		public static bool tExitFlag;	// command to totally exit the program
	}
}
