// Cryp.cs : C# driver for CrypF

//                                 Data Encryption
//                                  Fast & Secure
//
//                             ----------------------
//                      The algorithm for the Data Encryption
//                   Standard runs too slow on most micros, but
//                    simpler methods have not provided secure
//                      encryption. This program solves this
//                     problem by being both fast and secure.
//                             ----------------------
//
//                               By Harry J. Smith

using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Text;			// for Encoding I/O to a Text file

namespace MyNamespace
{
	/// <summary>
	///             CrypF, a data encryption system, Forms version
	///             Developed in TURBO Pascal 5.5 and 6.0, converted to C
	///             and now in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	public class Cryp
	{
		public Cryp()		// constructor logic
		{
			restore = false;
			nL = Environment.NewLine;
			Init();
		} // Cryp

		// ------- initialize/restart
		private static void Init()
		{
			file1 = "File1.txt";	// name of File1
			file2 = "File2.txt";	// name of File2
			fileC = "FileC.txt";	// name of FileC
			fileD = "FileD.txt";	// name of FileD
			fileE = "FileE.txt";	// name of FileE
			fileK = "FileK.txt";	// name of FileK
			fileX = "FileX.CFC";	// name of FileX
			repeat2 = true;
			param = "1";			// input key for encryption
			// Key24 = 1///////////////////////
			MultiID.WriteLn("Key = 1");
			GetKey();					// compute  keyStd
			Directory.SetCurrentDirectory(Common.homeDir);
			FileStream conFile = null;
			BinaryReader bC = null;
			if (OpenReadBr(out conFile, out bC, "Cry.con"))	// Try to open continuation file
			{
				int nr;
				// Read continuation file
				if (!ReadFBr(conFile, bC, cryCon, CONSIZE, out nr) || nr != CONSIZE)
					MultiID.WriteLn("Cannot read file \"Cry.con\"");
				conFile.Close();
			}
		} // init
		
  		public static void ChangeDir()	// change the default disk directory
		{
			string stL = '"' + param;
			Common.ChangeDir(ref stL);
		} // ChangeDir

		public static void Decode()	// Decode FileE to FileD using key
		{
			if (fileE == fileD)
			{
				MultiID.WriteLn("Cannot Decode file \"" + fileE + "\" to itself" + nL);
				return;
			}
			FileStream inF = null;
			BinaryReader bR = null;
			if (!OpenReadBr(out inF, out bR, fileE))	// Try to open FileE
				return;
			FileStream outF = null;
			BinaryWriter bW = null;
			if (!OpenWriteBw(out outF, out bW, fileD))	// Try to open FileD
			{
				inF.Close();
				return;
			}
			int nr;
			// Read first record of file
			if (!ReadFBr(inF, bR, buf, CONSIZE, out nr) || nr != CONSIZE)
			{
				MultiID.WriteLn("Cannot read file \"" + fileE + '"' + nL);
				outF.Close();
				inF.Close();
				return;
			}
			DoIt(inF, bR, outF, bW);
			outF.Close();
			inF.Close();
		} // Decode

		public static void Encode()	// Encode FileC to FileE using key
		{
			if (fileC == fileE)
			{
				MultiID.WriteLn("Cannot Encode file \"" + fileE + "\" to itself" + nL);
				return;
			}
			FileStream inF = null;
			BinaryReader bR = null;
			if (!OpenReadBr(out inF, out bR, fileC))	// Try to open FileC
				return;
			FileStream outF = null;
			BinaryWriter bW = null;
			if (!OpenWriteBw(out outF, out bW, fileE))	// Try to open FileE
			{
				inF.Close();
				return;
			}
			string savDir = Environment.CurrentDirectory.ToString();
			Directory.SetCurrentDirectory(Common.homeDir);		// Cry.con is always in homeDir
			FileStream conFile = null;
			BinaryReader bC = null;
			if (!OpenReadBr(out conFile, out bC, "Cry.con"))	// Try to open continuation file
			{
				outF.Close();
				inF.Close();
				Directory.SetCurrentDirectory(savDir);
				return;
			}
			Directory.SetCurrentDirectory(savDir);
			int nr;
			// Read continuation file
			if (!ReadFBr(conFile, bC, buf, CONSIZE, out nr) || nr != CONSIZE)
			{
				MultiID.WriteLn("Cannot read file \"Cry.con\"");
				conFile.Close();
				outF.Close();
				inF.Close();
				return;
			}
			conFile.Close();
			if (!WriteFBw(outF, bW, buf, nr))	// Write 1st record of fileE, if write error, exit
			{
				outF.Close();
				inF.Close();
				return;
			}
			DoIt(inF, bR, outF, bW);
			outF.Close();
			inF.Close();
			Term();			// Terminate processing, update Cry.con
		} // Encode

		public static void Equal(string fileA, string fileB)	// does the contents of fileA == fileB
		{
			string savDir = Environment.CurrentDirectory.ToString();
			if (savDir != Common.homeDir)
				Directory.SetCurrentDirectory(Common.homeDir);
			if (fileA == fileB && savDir == Common.homeDir)
			{
				MultiID.WriteLn("True, files \"" + fileA + "\" and \"" + fileB + "\" are the same file");
				return;
			}
			FileStream aFile = null;
			BinaryReader bA = null;
			if (!OpenReadBr(out aFile, out bA, fileA, true))  // Try to open FileA in homeDir only
			{
				if (savDir != Common.homeDir)
					Directory.SetCurrentDirectory(savDir);
				return;
			}
			if (savDir != Common.homeDir)
				Directory.SetCurrentDirectory(savDir);
			FileStream bFile = null;
			BinaryReader bB = null;
			if (!OpenReadBr(out bFile, out bB, fileB, true))  // Try to open FileB in CurrentDirectory first
			{
				aFile.Close();
				return;
			}
			if (aFile.Name == bFile.Name)
			{
				MultiID.WriteLn("True, files \"" + fileA + "\" and \"" + fileB + "\" are the same file");
				aFile.Close();
				bFile.Close();
				return;
			}
			if (aFile.Length != bFile.Length)
			{
				MultiID.WriteLn("False, files \"" + fileA + "\" and \"" + fileB + "\" are not the same length");
				aFile.Close();
				bFile.Close();
				return;
			}
			int i;
			int nA = -1;  // Number of bytes read from FileA
			int nB = -1;  // Number of bytes read from FileB
			while (nA != 0)
			{
				bool dOK = ReadFBr(aFile, bA, buf, BUFSIZE, out nA) && nA != 0;
				bool cOK = ReadFBr(bFile, bB, buf2, BUFSIZE, out nB) && nB != 0;
				if (dOK != cOK || nA != nB)
				{
					MultiID.WriteLn("False, files \"" + fileA + "\" and \"" + fileB + "\" are not the same length");
					aFile.Close();
					bFile.Close();
					return;
				}
				if (dOK && cOK)
				{
					for (i = 0; i < (nA + 1) / 2; i++)
					{
						if (buf[i] != buf2[i])
						{
							MultiID.WriteLn("False, files \"" + fileA + "\" and \"" + fileB + "\" are not equal");
							return;
						}
					}
				}
			}
			MultiID.WriteLn("True, files \"" + fileA + "\" and \"" + fileB + "\" are equal");
			aFile.Close();
			bFile.Close();
		} // Equal

		public static void DoIt(FileStream inF, BinaryReader bR, FileStream outF, BinaryWriter bW)
		// This is the heart of the encryption method
		{
			int i;
			for (i = 0; i < CONSIZE / 2; i++)
				pool[i] = (Int16)buf[i];
			for (i = 0; i < 9; ++i)			// Save and encode continuation seeds
			{
				rC[i] = pool[i];
				rI[i] = (Int16)(rS[i] ^ pool[i]);
				if (rI[i] == 0)				// Do not allow a zero seed
					rI[i] = 1;
			}
			rR = Rc9();						// Init first word of running
			next = 0;
			for (i = 0; i < POOLSIZE; ++i)	// Initialize pool, random int
			{
				pool[i] = rR += Rn(next);
				if (++next == 9)
					next = 0;
			}
			rR = Rc9();				// Init rR for first random index into pool
//			for (i = 0; i < 9; ++i)	// diag
//				MultiID.WriteLn("rS[" + i + "] = " + rS[i]);
//			for (i = 0; i < 9; ++i)
//				MultiID.WriteLn("rI[" + i + "] = " + rI[i]);
			int ind;		// Random index into pool of random integers
			int nr = -1;	// Number of bytes read into buf
			while (ReadFBr(inF, bR, buf, BUFSIZE, out nr) && nr != 0)
			{
				for (i = 0; i < (nr + 1) / 2; i++)
				{
					ind = rR & poolMask;			// Random index to random pool
					buf[i] = (UInt16)(buf[i] ^ pool[ind]);	// XOR data with random int
					pool[ind] ^= ( rR += Rn(next));	// XOR random int with next rR
					if( ++next == 9)				// Cycle to next generator
						next = 0;
				}
				if (!WriteFBw(outF, bW, buf, nr))	// Write output file, if write error, exit
					return;
			}
		} // DoIt

		public static void Term()	// Terminate processing, update Cry.con
		{
			string savDir = Environment.CurrentDirectory.ToString();
			Directory.SetCurrentDirectory(Common.homeDir);
			FileStream conFile = null;
			BinaryWriter bW = null;
			if (OpenWriteBw(out conFile, out bW, "Cry.con"))
			{
				int i;
				for (i = 0; i < 9; ++i)			// Restore clear continuation seeds
					rI[i] = rC[i];
				for (i = POOLSIZE - 1; i >= 0; i--)	// Build a random record
					pool[i] = Rc9();
				for (i = 0; i < CONSIZE / 2; i++)	// copy to buf
					cryCon[i] = (UInt16)pool[i];
				// Write continuation file
				if (WriteFBw(conFile, bW, cryCon, CONSIZE))
					if (!Common.quiet)
						MultiID.WriteLn("Continuation file \"Cry.con\" updated");
				conFile.Close();
			}
			Directory.SetCurrentDirectory(savDir);
		} // Term

		public static void File1()	// Set Input file name of file1 for Equal12 command
		{
			file1 = param;	// name of File1
			if (!restore)
				VerifyF();
		} // File1

		public static void File2()	// Set Input file name of file2 for Equal12 command
		{
			file2 = param;	// name of File2
			if (!restore)
				VerifyF();
		} // File2

		public static void FileC()	// Set Input file name of clear file to be encrypted
		{
			fileC = param;	// name of FileC
			if (!restore)
				VerifyF();
		} // FileC

		public static void FileD()	// Set Output file name of file to decrypt to
		{
			fileD = param;	// name of FileD
			if (!restore)
				VerifyF();
		} // FileD

		public static void FileE()	// Set Output file name of file to encrypt to
		{
			fileE = param;	// name of FileE
			if (!restore)
				VerifyF();
		} // FileE

		public static void FileK()	// Set file name of key file
		{
			fileK = param;	// name of FileK
			if (!restore)
				VerifyF();
		} // FileK

		public static void FileL()	// Set file name of Log file
		{
			Calc.cFLOut = param;
			if (!restore)
				VerifyF();
		} // FileL

		public static void FileX()	// Set file name of eXecute file
		{
			fileX = param;	// name of FileX
			if (!restore)
				VerifyF();
		} // FileX

		public static void VerifyF()	// verify that File param exists
		{
			FileStream inF = null;
			BinaryReader bR = null;
			if (OpenReadBr(out inF, out bR, param))		// Try to open new File param
			{
				inF.Close();
				if (!Common.quiet)
				{
					MultiID.WriteLn("File \"" + param + "\" verified and closed");
					MultiID.WriteLn("");
				}
			}
		} // VerifyF

		public static void InitC()	// Initialize the Continuation File Cry.con
		{
			string savDir = Environment.CurrentDirectory.ToString();
			Directory.SetCurrentDirectory(Common.homeDir);
			FileStream conFile = null;
			BinaryWriter bW = null;
			if (OpenWriteBw(out conFile, out bW, "Cry.con"))
			{
				int i;
				for (i = 0; i < 9; i++)				// Initialize Cry.con file
					rI[i] = rS[i];
				for (i = POOLSIZE - 1; i >= 0; i--)	// Build a random record
					pool[i] = Rc9();
				for (i = 0; i < CONSIZE / 2; i++)	// copy to buf
					cryCon[i] = (UInt16)pool[i];
				// Write continuation file
				if (WriteFBw(conFile, bW, cryCon, CONSIZE))
					if (!Common.quiet)
						MultiID.WriteLn("Continuation file \"Cry.con\" updated" + nL);
				conFile.Close();
			}
			Directory.SetCurrentDirectory(savDir);
		} // InitC

		public static void GetKey()	// Input a key from keyboard
		{
			keyIn = param;
			StdKey();
			MultiID.WriteLn("");
		} // GetKey

		public static void GetKeyF()	// Input a key from FileK
		{
			StreamReader inF = null;
			if (!OpenReadSt(out inF, fileK))	// Try to open key file
				return;
			keyIn = inF.ReadLine().Trim();		// get the first line of fileK
			inF.Close();
			MultiID.WriteLn("Key = " + keyIn);
			StdKey();
			MultiID.WriteLn("");
		}
		// GetKeyF

		public static void StdKey()	// Standardize keyIn
		{
			int i, j;	// Utility indexes
			char ch;	// Utility character
			string key;	// Temp value of key
			keyStd = keyIn;
			Standard(ref keyStd);
			key = keyStd;
			int len = keyIn.Length;
			if (len > MAXKEY)
			{
				for (i = 1; i < MAXKEY; i++)
				{
					ch = (char)((int)(key[i]) + (int)(key[i - 1]));
					SetSt(ref key, i, ch);
				}
				j = len - 1;
				for (i = 25; i < len; i++)
				{
					j = j - 1;
					ch = (char)((int)(key[j]) + (int)(key[j + 1]));
					SetSt(ref key, j, ch);
				}
			}
			key24 = "";
			for (i = 0; i < MAXKEY; i++)
				key24 += (char)(0);
			j = 0;
			for (i = 0; i < len; i++)
			{
				ch = (char)((int)(key24[j]) + (int)(key[i]));
				SetSt(ref key24, j, ch);
				Math.DivRem(j + 1, MAXKEY, out j);
			}
			for (i = 0; i < MAXKEY; i++)
				if (key24[i] == (char)(0))
					SetSt(ref key24, i, ' ');
			Standard(ref key24);
			ConvKey();
			MultiID.WriteLn(keyStd + " = input key in standard form");
			MultiID.WriteLn(key24 + " = 24-character key in standard form");
		} // StdKey

		public static void ConvKey()	// Convert the encryption key to 9 seeds
		// Convert the encryption key to 9 seeds
		//
		// input: key24[i], i = 0, ..., 23
		// output:   rS[j], j = 0, ..., 8
		//
		//   [0]    [1]    [2]    [3]    [4]    [5]    [6]   ...  [22]   [23]
		// 111111 111111 111111 111111 111111 111111 111111      111111 111111
		//    6      6     4  2    6      6    2  4     6       6      6
		// ****** seed0 *****...... seed1 ......***** seed2  ...  seed8 ******
		//       16 BITS           16 BITS       16 BITS     16 BITS
		//
		// 144 bit key converted to 144 bits of seed,
		// seeds of zero are changed to 1.
		{
			int  i, j;
			int[] keyI = new int[MAXKEY];	// Key in integer form, 6 bits each int
			for(i = 0; i < MAXKEY; i++)
				keyI[i] = key24[i] & 63;
			for(i = 0, j = 0; j < 9; i += 8, j += 3)
			{
				rS[j] =	(Int16)((keyI[i]   << 10) |
						(keyI[i+1] <<  4) |
						(keyI[i+2] >>  2));

				rS[j + 1] = (Int16)((keyI[i + 2] << 14) |
							 (keyI[i + 3] << 8) |
							 (keyI[i + 4] << 2) |
							 (keyI[i + 5] >> 4));

				rS[j + 2] = (Int16)((keyI[i + 5] << 12) |
							 (keyI[i + 6] << 6) |
							  keyI[i + 7]);
			}
			for(j = 0; j < 9; j++)		// Do not allow a zero seed
				if(rS[j] == 0)
					rS[j] = 1;
			repeat2 = true;
		} // ConvKey

		public static void Standard(ref string key)	// Standardize a key string
		// Output characters are from 33 to 95 inclusive
		{
			char ch;
			int len = key.Length;
			for (int i = 0; i < len; i++)
			{
				ch = key[i];
				ch &= (char)127;
				if (ch > 95)
					ch -= (char)32;
				if (ch < 32)
					ch += (char)32;
				if (ch == ' ')
					ch = '/';
				SetSt(ref key, i, ch);	// key[i] = ch;
			}
		} // Standard

		public static void Log()	// Set Log screen to log file mode on or off
		{
			bool loggL = param != "0";
			if (MultiID.logg != loggL)
				Common.ToggleLogg();
		} // Log

		public static void LogT()	// Set Log screen to log file mode on (True)
		{
			if (!MultiID.logg)
				Common.ToggleLogg();
		} // LogT

		public static void LogF()	// Set Log screen to log file mode off (False)
		{
			if (MultiID.logg)
				Common.ToggleLogg();
		} // LogF

		public static void Quiet()	// Set Quiet mode on or off
		{
			Common.quiet = param != "0";
		} // Quiet

		public static void QuietT()	// Set Quiet mode on (True)
		{
			Common.quiet = true;
		} // QuietT

		public static void QuietF()	// Set Quiet mode off (False)
		{
			Common.quiet = false;
		} // QuietF

		public static void Save()	// Save the state of inputs to Save.CFC
		{
			StreamWriter outF = null;
			if (!OpenWriteSt(out outF, "Save.CFC"))
				return;
			if (!WriteFSt(outF, "File1 = " + QuoteFN(file1) + nL))
			{
				outF.Close();
				return;
			}
			WriteFSt(outF, "File2 = " + QuoteFN(file2) + nL);
			WriteFSt(outF, "FileC = " + QuoteFN(fileC) + nL);
			WriteFSt(outF, "FileD = " + QuoteFN(fileD) + nL);
			WriteFSt(outF, "FileE = " + QuoteFN(fileE) + nL);
			WriteFSt(outF, "FileK = " + QuoteFN(fileK) + nL);
			WriteFSt(outF, "FileX = " + QuoteFN(fileX) + nL);
			WriteFSt(outF, "Key = " + keyIn + nL);
			outF.Close();
			MultiID.WriteLn("Save file \"Save.CFC\" updated" + nL);
		} // Save

		public static string QuoteFN(string fileName)	// quote fileName if needed
		{
			if (fileName.Contains(" "))
				return '"' + fileName + '"';
			for (int i = 0; i < Common.delim.Length; i++)
			{
				if (fileName.Contains(Common.delim[i].ToString()))
					return '"' + fileName + '"';
			}
			return fileName;
		} // QuoteFN

		public static void State()	// Show state of inputs
		{
			MultiID.WriteLn(" Home Dir: \"" + Common.homeDir + '"');
			MultiID.WriteLn("File path: \"" + Environment.CurrentDirectory.ToString() + '"');
			MultiID.WriteLn("");
			MultiID.WriteLn(" File1 = " + QuoteFN(file1));
			MultiID.WriteLn(" File2 = " + QuoteFN(file2));
			MultiID.WriteLn(" FileC = " + QuoteFN(fileC));
			MultiID.WriteLn(" FileD = " + QuoteFN(fileD));
			MultiID.WriteLn(" FileE = " + QuoteFN(fileE));
			MultiID.WriteLn(" FileK = " + QuoteFN(fileK));
			MultiID.WriteLn(" FileX = " + QuoteFN(fileX));
			MultiID.WriteLn(" KeyIn = " + keyIn);
			MultiID.WriteLn("KeyStd = " + keyStd);
			MultiID.WriteLn(" Key24 = " + key24);
			MultiID.WriteLn("");
		} // State

		public static bool OpenReadBr(out FileStream inF, out BinaryReader bR, string fName)
		// open file for Reading in CurrentDirect or homeDir
		{
			return OpenReadBr(out inF, out bR, fName, false);
		} // OpenReadBr

		public static bool OpenReadBr(out FileStream inF, out BinaryReader bR, string fName, bool currOnly)
		// open file for Reading in CurrentDirect or homeDir
		{
			string stEx1;
			string stEx2 = "";
			string savDir = Environment.CurrentDirectory.ToString();
			inF = null;	// inF must be assigned to before control leaves the current method
			bR = null;	// bR must be assigned to before control leaves the current method
			if (OpenReadFileBr(out inF, out bR, fName, out stEx1))
			{
				if (!Common.quiet)
				{
					MultiID.WriteLn("File \"" + fName + "\" opened for reading");
					Common.WCurDir(fName);
				}
				return true;
			}
			if (savDir != Common.homeDir && !currOnly)
				if (savDir != Common.homeDir)
				{
					Directory.SetCurrentDirectory(Common.homeDir);
					if (OpenReadFileBr(out inF, out bR, fName, out stEx2))
					{
						if (!Common.quiet)
						{
							MultiID.WriteLn("File \"" + fName + "\" opened for reading");
							Common.WCurDir(fName);
						}
						Directory.SetCurrentDirectory(savDir);
						return true;
					}
					Directory.SetCurrentDirectory(savDir);
				}
			MultiID.WriteLn("Cannot open file \"" + fName + "\" for input");
			MultiID.WriteTLn("Not in current directory \"" + savDir + '"');
			Common.ReportError("OpenRead", stEx1);           // file not found most likely
			if (savDir != Common.homeDir && !currOnly)
			{
				MultiID.WriteTLn("Not in home directory \"" + Common.homeDir + '"');
				Common.ReportError("OpenRead", stEx2); // file not found most likely
			}
			return false;
		} // OpenReadBr

		public static bool OpenReadFileBr(out FileStream inF, out BinaryReader bR, string fName, out String stEx)
		// open file for Reading from CurrentDirectory
		{
			inF = null;	// inF must be assigned to before control leaves the current method
			bR = null;	// bR must be assigned to before control leaves the current method
			stEx = "";
			try
			{
				inF = new FileStream(fName, FileMode.Open);	// open fName for input as inF
				bR = new BinaryReader(inF);
				return true;
			}
			catch (Exception ex)
			{
				stEx = ex.Message.ToString();
				return false;
			}
		} // OpenReadFileBr

		public static bool OpenWriteBw(out FileStream outF, out BinaryWriter bW, string fName)
		// open file for Writing to CurrentDirect
		{
			outF = null;	// outF must be assigned to before control leaves the current method
			bW = null;		// bW must be assigned to before control leaves the current method
			try
			{
				if (File.Exists(fName))
					File.Delete(fName);
				// open fName for output as outF
				outF = new FileStream(fName, FileMode.CreateNew);	// Appending = false
				bW = new BinaryWriter(outF);
				if (!Common.quiet)
				{
					MultiID.WriteLn("File \"" + fName + "\" opened for writing");
					Common.WCurDir(fName);
				}
				return true;
			}
			catch (Exception ex)
			{
				MultiID.WriteLn("Cannot open file \"" + fName + "\" for output");
				MultiID.WriteLn("OpenWrite: " + ex.Message + nL);
				return false;
			}
		} // OpenWriteBw

		public static bool OpenWriteSt(out StreamWriter outF, string fName)
		// open file for Writing to CurrentDirect
		{
			outF = null;	// outF must be assigned to before control leaves the current method
			try
			{
				// open fName for output as outF
				outF = new StreamWriter(fName, false, Encoding.ASCII);	// Appending = false
				if (!Common.quiet)
				{
					MultiID.WriteLn("File \"" + fName + "\" opened for writing");
					Common.WCurDir(fName);
				}
				return true;
			}
			catch (Exception ex)
			{
				MultiID.WriteLn("Cannot open file \"" + fName + "\" for output");
				MultiID.WriteLn("OpenWrite: " + ex.Message + nL);
				return false;
			}
		} // OpenWriteSt

		public static bool WriteFBw(FileStream outF, BinaryWriter bW, UInt16[] buf, int n)
		// write n bytes from buf to an opened file
		{
			if (buf.Length == 0 || n == 0)
				return true;
			if (outF == null || bW == null)
				return false;
			if (n > 2 * buf.Length)
				n = 2 * buf.Length;
			byte b;
			try
			{
				for (int i = 0; i < n; i++)
				{
					if ((i & 1) == 0)
						b = (byte)buf[i / 2];
					else
						b = (byte)(buf[i / 2] >> 8);
					bW.Write(b);
				}
				return true;
			}
			catch (Exception ex)
			{
				MultiID.WriteLn("File write error " + ex.Message + nL);
				return false;
			}
		} // WriteFBw

		public static bool ReadFBr(FileStream inF, BinaryReader bR, UInt16[] buf, int n, out int nr)
		// read upto n bytes from an opened file to buf
		{
			nr = 0;
			if (buf.Length == 0 || n == 0)
				return false;
			if (inF == null || bR == null)
				return false;
			if (n > 2 * buf.Length)
				n = 2 * buf.Length;
			nr = 0;
			byte b;
			try
			{
				for (int i = 0; i < n; i++)
				{
					b = bR.ReadByte();
					if ((i & 1) == 0)
						buf[i / 2] = b;
					else
						buf[i / 2] = (UInt16)(buf[i / 2] + (b << 8));
					nr++;
				}
				return true;
			}
			catch { return true; }
		} // ReadFBr

		public static bool WriteFSt(StreamWriter outF, string st)	// write string to an opened file
		{
			if (st == "")
				return false;
			if (outF == null)
				return false;
			try
			{
				outF.Write(st);
				return true;
			}
			catch (Exception ex)
			{
				MultiID.WriteLn("File write error " + ex.Message + nL);
				return false;
			}
		} // WriteFSt

		public static bool OpenReadSt(out StreamReader inF, string fName)
		// open file for Reading from CurrentDirectory or homeDir
		{
			string dir;
			return Common.OpenRead(out inF, fName, out dir);
		} // OpenReadSt

		public static void SetSt(ref string st, int i, char ch)	// set st[i] = ch
		{
			int len = st.Length;
			if (i < 0 || i >= len)
				return;
			st = st.Substring(0, i) + ch + st.Substring(i + 1, len - i - 1);
		} // SetSt

		public static Int16 Rn(int n)	// Returns the next random integer from rI[n], 0 <= n <= 8
		{
			int i;
			switch(n)
			{
			case 0:
				// 1st Congruential Generator, 16 bits
				// Generates Random integers from -32768 to 32767 inclusive
				// Cycle length = 65536 = 2**16
				//
				return (rI[0] = (Int16)(25173 * rI[0] + 6925));
			case 1:
				//  1st Shift-register Generator, 16 bits
				//  Random integer from -32768 to 32767 inclusive, 0 not generated
				//  Generator = -22620, Cycle length = 65535 = 3 * 5 * 17 * 257
				//
				if((rI[1] & 1) == 1)	// if odd
					return (rI[1] = (Int16)(((UInt16)rI[1] >> 1) ^ -22620));
				return (rI[1] = (Int16)((UInt16)rI[1] >> 1));
			case 2:
				//  2nd Congruential Generator, 16 bits
				//  Random integer from -32768 to 32767 inclusive
				//  Cycle length = 65537 = prime, zero repeats once
				//
				if (rI[2] == 0)
				{
					if (repeat2)
					{
						repeat2 = false;
						return 0;
					}
					else repeat2 = true;
				}
				return (rI[2] = (Int16)(23629 * rI[2] + 13849));
			case 3:
				//  2nd Shift-register Generator, 16 bits
				//  Random integer from -32768 to 32767, not all generated
				//  Generator = -07493, Cycle length = 65521 = prime, (65536 - 15)
				//
				if((rI[3] & 1) == 1)	// if odd
					rI[3] = (Int16)(((UInt16)rI[3] >> 1) ^ -7493);
				else
					rI[3] = (Int16)((UInt16)rI[3] >> 1);
				if (rI[3] == 1)
					for (i = 0; i < 14; i++)
						Rn(3);			// Throw 14 away
				return rI[3];
			case 4:
				//  3rd Congruential Generator, 16 bits
				//  Random integer from -32768 to 32767, not all generated
				//  Cycle length = 65519 = prime, (65536 - 17)
				//
				rI[4] = (Int16)(4821 * rI[4] + 13001);
				if (rI[4] == 0)
					for (i = 0; i < 17; i++)
						Rn(4);			// Throw 17 away
				return rI[4];
			case 5:
				//  3rd Shift-register Generator, 16 bits
				//  Random integer from -32768 to 32767, not all generated
				//  Generator = -25501, Cycle length = 65497 = prime, (65536 - 39)
				//
				if ((rI[5] & 1) == 1)
					rI[5] = (Int16)(((UInt16)rI[5] >> 1) ^ -25501);
				else
					rI[5] = (Int16)((UInt16)rI[5] >> 1);
				if (rI[5] == 1)
					for (i = 0; i < 38; i++)
						Rn(5);			// Throw 38 away
				return rI[5];
			case 6:
				//  4th Congruential Generator, 16 bits
				//  Random integer from -32768 to 32767, not all generated
				//  Cycle length = 65479 = prime, (65536 - 57)
				//
				rI[6] = (Int16)(10349 * rI[6] + 7001);
				if (rI[6] == 0)
					for (i = 0; i < 57; i++)
						Rn(6);			// Throw 57 away
				return rI[6];
			case 7:
				//  4th Shift-register Generator, 16 bits
				//  Random integer from -32768 to 32767, not all generated
				//  Generator = -18916, Cycle length = 65449 = prime, (65536 - 87)
				//
				if ((rI[7] & 1) == 1)
					rI[7] = (Int16)(((UInt16)rI[7] >> 1) ^ -18916);
				else
					rI[7] = (Int16)((UInt16)rI[7] >> 1);
				if (rI[7] == 1)
					for (i = 0; i < 86; i++)
						Rn(7);			// Throw 86 away
				return rI[7];
			case 8:
				//  5th Congruential Generator, 16 bits
				//  Random integer from -32768 to 32767, not all generated
				//  Cycle length = 65447 = prime, (65536 - 89)
				//
				rI[8] = (Int16)(30133 * rI[8] + 14001);
				if(rI[8] == 0)
					for (i = 0; i < 89; i++)
						Rn(8);			// Throw 89 away
				return rI[8];
			}
			return 0;					// This line is never reached
		} // Rn

		public static Int16 Rc9()		// Combination of Rn(0) thru Rn(8), one each
		{
			int i;
			Int16 rC = 0;
			for (i = 0; i < 9; i++)
				rC += Rn(i);
			return rC;
		} // Rc9

		// constants
		private static int BUFSIZE = 16384;	// Number of characters/bytes in file i/o buffer
		private static int CONSIZE = 18;	// Number of characters/bytes in Cry.con file
		private static int MAXKEY   = 24;	// Number of characters/bytes in short key
		private static int POOLSIZE = 2048;	// Number of 16-bit integers in random number pool

		private static bool repeat2;	// repeat flag used in Rn(2)
		public static bool restore;	// Restore flag
		private static int next;		// Next random no. routine to use, cycles 0 thru 8
		private static int poolMask = POOLSIZE - 1;	// Mask for index to random pool = 2**N - 1
		private static Int16 rR;		// Random running combination of rI[0] thru rI[8]
		private static Int16[] pool = new Int16[POOLSIZE];	// Pool of POOLSIZE random integers
		private static Int16[] rC = new Int16[9];	// Current value of clear continuation seeds
		private static Int16[] rI = new Int16[9];	// Current random integers from the 9 generators
		private static Int16[] rS = new Int16[9];	// Initial random seeds for the 9 generators
		public static string file1;		// Input file name of file1 for Equal12 command
		public static string file2;		// Input file name of file2 for Equal12 command
		public static string fileC;		// name of FileC, clear file to be encrypted
		public static string fileD;		// name of FileD, file to decrypt to
		public static string fileE;		// name of FileE, file to encrypt to or decrypt
		public static string fileK;		// name of FileK, key file
		public static string fileX;		// name of FileX, eXecute file
		public static string key24;		// 24-character key in standard form
		public static string keyIn;		// input key for encryption
		public static string keyStd;	// input key in standard form
		private static string nL;		// one Environment.NewLine \n or \r or \r\n
		public static string param;		// parameter in command string
		private static UInt16[] buf = new UInt16[BUFSIZE / 2];		// Buffer for file i/o
		private static UInt16[] buf2 = new UInt16[BUFSIZE / 2];		// Buffer for file i/o
		public static UInt16[] cryCon = new UInt16[CONSIZE / 2];	// Buffer for Cry.con file i/o
	}
}
