#undef  SPEEDUP				// 0 // 0 or 1 (default is 1)
#define SPEEDUP				// 1 // 0 or 1 (default is 1)

using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Windows.Forms; // for RichTextBox

namespace MyNamespace
{
	/// <summary>
	///		MultiID part of CrypF, a data encryption system, Forms version
	///		from my XICalc - Multiple precision integer algorithms
	///		Developed in TURBO Pascal 5.5 and 6.0, converted to Visual Basic 6
	///		and now in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	public class MultiID
	{
		public MultiID() // constructor logic
		// initialize Multi-precision integer package
		// input: muDMax, muNMax, baseM
		{
			muErr = false;
			muErrRep = true;
			muAbort = false;
			statReq = false;
			statReq2 = false;
			statReq3 = false;
			time0 = timeP = Environment.TickCount/1000.0;
			tV1 = tV2 = 0;
			nL = Environment.NewLine;
			nL2 = nL + nL;
			nLc = nL;					// line continuation mark for WriteT nL or "-" + nL
			muICount = 0;				// belongs to MuInterrupt()
			muITimeSaved = 0;			// belongs to MuInterrupt()
			diagOn = false;
			debug = false;
		}

		public static void Write(string message) // write a short message to outS
		{
			Write(outS, null, message);
		}

		public static void WriteLn() // write a linefeed to outS
		{
			WriteLn(outS, null, "");
		}

		public static void WriteLn(string message) // write a short message and a linefeed to outS
		{
			WriteLn(outS, null, message);
		}

		public static void WriteLn(RichTextBox outB, StreamWriter outF) // Write linefeed
		{
			Write(outB, outF, nL);
		}

		public static void WriteLn(RichTextBox outB, StreamWriter outF, string st) // Write st plus linefeed
		{
			Write(outB, outF, st + nL);
		}

		public static void WriteLn2() // Write 2 linefeeds to outS
		{
			Write(outS, null, nL2);
		}

		public static void WriteLn2(string st) // Write st plus 2 linefeeds to outS
		{
			Write(outS, null, st + nL2);
		}

		public static void WriteLn2(RichTextBox outB, StreamWriter outF) // Write 2 linefeeds
		{
			Write(outB, outF, nL2);
		}

		public static void WriteLn2(RichTextBox outB, StreamWriter outF, string st) // Write st plus 2 linefeeds
		{
			Write(outB, outF, st + nL2);
		}

		public static void WriteSLn(string st) // Write SingleSpace, st, and linefeed to outS
		{
				Common.SingleSpace(outS);
				Write(outS, null, st + nL);
		}

		public static void WriteDLn(string st) // Write DoubleSpace, st, and linefeed to outS
		{
				Common.DoubleSpace(outS);
				Write(outS, null, st + nL);
		}

		public static void WriteDLn2(string st) // Write DoubleSpace, st, and 2 linefeeds to outS
		{
				Common.DoubleSpace(outS);
				Write(outS, null, st + nL2);
		}

		public static void WriteDLn(RichTextBox outB, StreamWriter outF, string st) // Write DoubleSpace, st, and linefeed
		{
				Common.DoubleSpace(outB);
				Write(outB, outF, st + nL);
		}

		public static void Write(RichTextBox outB, StreamWriter outF, string st) // print string to RichTextBox or to a file
		{
			if (st == "") return;
			if (outF != null)
			{
				try	{outF.Write(st);}
				catch (Exception ex) {Common.ReportError("Write", ex.Message);}
			}
			if (outB != null)
			{
				outB.AppendText(st);
				if (st[(st.Length - 1)] == '\n')	// if a new line, move to bottom/end of output text box
				{
					Global.runTextOut.Select(Global.runTextOut.Text.Length, 0);
					Global.runTextOut.Focus();
				}
				if (logg && (outB == outS))			// only log output to Run form textbox
				{
					try	{outL.Write(st);}
					catch {}
				}
				if (outB.Text.Length > 200000)		// clear top of screen after about 50 full pages
				{
					int i = outB.Text.IndexOf("\n", 4000);
					outB.Text = outB.Text.Substring(i + 1);	// delete partial line
				}
			}

		} // Write

		public static void WriteT(string st)
		// WriteT with 0 tot and 0 width to outS
		{
			int tot = 0;
			Common.SingleSpace();
			WriteT(outS, null, st, ref tot, 0);
		}

		public static void WriteT(string st, ref int tot)
		// WriteT with 0 width to outS
		{
			WriteT(outS, null, st, ref tot, 0);
		}

		public static void WriteT(RichTextBox outB, StreamWriter outF, string st)
		// WriteT with 0 tot and 0 width
		{
			int tot = 0;
			Common.SingleSpace(outB);
			WriteT(outB, outF, st, ref tot, 0);
		}

		public static void WriteT(RichTextBox outB, StreamWriter outF, string st, ref int tot)
		// WriteT with 0 width
		{
			WriteT(outB, outF, st, ref tot, 0);
		}

		public static void WriteTLn()
		// WriteTLn with "", 0 tot and 0 width to outS
		{
			Write(outS, null, nL);
		}

		public static void WriteTLn(string st)
		// WriteTLn with 0 tot and 0 width to outS
		{
			int tot = 0;
			Common.SingleSpace();
			if (st != "") WriteT(outS, null, st, ref tot, 0);
			Write(outS, null, nL);
		}

		public static void WriteTLn(string st, ref int tot)
		// WriteT and linefeed to outS
		{
			if (st != "") WriteT(outS, null, st, ref tot, 0);
			Write(outS, null, nL);
			tot = 0;
		}

		public static void WriteTLn(RichTextBox outB, StreamWriter outF, string st)
		// WriteT and linefeed with 0 tot and 0 width
		{
			int tot = 0;
			Common.SingleSpace(outB);
			if (st != "") WriteT(outB, outF, st, ref tot, 0);
			Write(outB, outF, nL);
		}

		public static void WriteTLn(RichTextBox outB, StreamWriter outF, string st, ref int tot)
		// WriteT and linefeed with 0 width
		{
			if (st != "") WriteT(outB, outF, st, ref tot, 0);
			Write(outB, outF, nL);
			tot = 0;
		}

		public static void WriteTLn(RichTextBox outB, StreamWriter outF, string st, ref int tot, int width)
		// WriteT and linefeed
		{
			if (st != "") WriteT(outB, outF, st, ref tot, width);
			Write(outB, outF, nL);
			tot = 0;
		}

		public static void WriteTLn2(string st)
		// WriteTLn2 with 0 tot and 0 width to outS
		{
			int tot = 0;
			Common.SingleSpace();
			if (st != "") WriteT(outS, null, st, ref tot, 0);
			Write(outS, null, nL2);
		}

		public static void WriteTLn2(RichTextBox outB, StreamWriter outF, string st)
		// WriteTLn2 with 0 tot and 0 width
		{
			int tot = 0;
			Common.SingleSpace(outB);
			if (st != "") WriteT(outB, outF, st, ref tot, 0);
			Write(outB, outF, nL2);
		}

		public static void WriteTLn2(RichTextBox outB, StreamWriter outF, string st, ref int tot, int width)
		// WriteT and a double linefeed after
		{
			if (st != "") WriteT(outB, outF, st, ref tot, width);
			Write(outB, outF, nL2);
			tot = 0;
		}

		public static void WriteTD(RichTextBox outB, StreamWriter outF, string st)
		// force double space and WriteT with 0 tot and 0 width
		{
			int tot = 0;
			Common.SingleSpace(outB);
			WriteT(outB, outF, st, ref tot, 0);
		}

		public static void WriteTDLn(string st)
		// WriteT with a guaranteed double linefeed before and single linefeed after to outS
		{
			Common.DoubleSpace(outS);
			WriteTLn(outS, null, st);
		}

		public static void WriteTDLn(RichTextBox outB, StreamWriter outF, string st)
		// WriteT with a guaranteed double linefeed before and single linefeed after
		{
			Common.DoubleSpace(outB);
			WriteTLn(outB, outF, st);
		}

		public static void WriteT(RichTextBox outB, StreamWriter outF, string st, ref int tot, int width)
		// output String and linefeed every 60, 80 or width characters
		// tot = characters on line so far
		{
			int i;
			int j;
			int iL;
			int n;
			string stE;
			if (width > 0)
				n = width;
			else
			{
				if (outB == null) n = 60;	// 80 for screen, 60 for files
				else n = 80;
			}
			i = 1;
			j = n - tot;
			iL = st.Length;
			int ix = 0;
			while (iL > 0)
			{
				stE = "";
				if (tot >= n)
				{
					stE = nLc;				// format n characters per line
					j = n;
					tot = 0;
				}
				if (iL <= j)
				{
					Write(outB, outF, stE + st.Substring(i - 1));	// built for speed, don't call Write too often
					ix = st.Substring(i - 1).LastIndexOf("\n");
					tot += iL;
					if (ix >= 0)
						tot = st.Substring(i - 1).Length - ix - 1;
					iL = 0;
				}
				else
				{
					Write(outB, outF, stE + st.Substring(i-1, j));
					ix = st.Substring(i - 1, j).LastIndexOf("\n");
					tot = n;
					if (ix >= 0)
						tot = st.Substring(i - 1, j).Length - ix - 1;
					i += j;
					iL -= j;
					j = n;
				}
			}
		} // WriteT

		public static void MuWriteErr(string st) // write error statement to outS and set error flag
		{
			muErr = true;
			if (muErrRep) WriteDLn(st);
		}

		public static void DiagT(string message, ref int tot) // output a diagnostic message to outS
		{
			if (diagOn) WriteT(message, ref tot);
		}

		public static void ADiag(string message) // output a diagnostic message to outS
		{
			if (diagOn) WriteLn(message);
		}

		public static void ADebug(string message) // output a debug message to outS
		{
			if (debug) WriteLn(message);
		}

		public static void ADDiag(string message) // output DoubleSpace and diagnostic message to outS
		{
			if (diagOn) WriteDLn(message);
		}

		public static void MuInterrupt() // test for ESC key pressed to abort operation
		// muAbort set true if Abort selected, unchanged if not
		{
			muICount++;
			if (muICount > 100)
			{
				muICount = 0;
				int Time = Environment.TickCount;	// time in milliseconds
				if (Time - muITimeSaved > 500 || Time < muITimeSaved)
				{
					// don't hog the processor, but don't slowdown too much
					Application.DoEvents();		// process user requested actions;
					muITimeSaved = Time;
				}
			}
			if (!abortCalc) return;
			if (muAbort) return;
			double TimeSaved = TimerF();	// allow for time paused
			DialogResult MsgButton = MessageBox.Show("*** INTERRUPT: To Continue select Ignore or Retry\n" +
									"To Abort Computation select Abort\n",
				Global.title,
				MessageBoxButtons.AbortRetryIgnore,
				MessageBoxIcon.Information);
			// ----------
			double TimeOut = TimerF() - TimeSaved;	// allow for time-out
			time0 += TimeOut;
			abortCalc = false;
			if (MsgButton == DialogResult.Ignore || MsgButton == DialogResult.Retry)
				WriteDLn("Continuing...");
			if (MsgButton == DialogResult.Abort)
			{
				muAbort = true;
				WriteDLn("Computation aborted by operator!");
			}
			RunForm.FocusBothUs();
		} // MuInterrupt

		public static double TimerF() // get the value of the system clock in total seconds from start of run
		// good to one clock tick ~= 0.0115 seconds (0.055 on some systems)
		// must call once each 24 days to maintain time0
		{
			double time1 = Environment.TickCount/1000.0; // TickCount = number of milliseconds since Windows started.
			if (time1 < timeP)							 // TickCount cycles 2^31 mills = 24.855... days
				time0 -= Math.Pow(2, 31)/1000;			 // time0 = 2^53 milliseconds = 285,420.9... years (long enough for me)
			timeP = time1;
			return time1 - time0;
		}

		public static void TimeDel() // output delta time only
		{
			int tot = 0;
			TimeDel(ref tot);
		}

		public static void TimeDel(ref int tot) // output delta time only
		{
			GC.Collect();
			long totalMemory = GC.GetTotalMemory(false);
			if (timeIt)
			{
				string st = ", Memory in use = " + totalMemory + ", DT = ";
				double tV2L = TimerF();
				double delta = tV2L - tV2;	// delta time from last TimeIt call
				WriteTLn(st + delta.ToString("0.00") + " sec.", ref tot);
			}
			else
			{
				WriteTLn("", ref tot);
			}
		}

		public static void TimeIt(string message) // output a diagnostic message with delta times
		{
			if (timeIt)
			{
				tV1 = tV2;
				tV2 = TimerF();
				double delta = tV2 - tV1;	// delta time from last TimeIt call
				WriteSLn("T = " + tV2.ToString("0.00") +
					"  DT = " + delta.ToString("0.00") + " sec.  " + message);
			}
		}

		public static void Diag(string message) // output a diagnostic message with delta times
		{
			if (diagOn) TimeIt(message);
		}

		public static void Delete(ref string st, int index, int count)
		// delete count characters starting at index, count >= 0
		{
			st = st.Substring(0, index) + st.Substring(index + count);
		}

		private static int muICount;		// belongs to MuInterrupt()
		private static int muITimeSaved;	// belongs to MuInterrupt()

		public static bool abortCalc;	// true if abort calc button clicked
		public static bool debug;		// debug mode turned on
		public static bool diagOn;		// diagnostics turned on
		public static bool logg;		// true if Logging screen output to FILE
		public static bool muAbort;		// true if abort calc confirmed by operator
		public static bool muErr;		// true if multi-precision error occurred
		public static bool muErrRep;	// true if multi-precision error reports are desired
		public static bool statReq;		// true if Status of calculation is requested, Level 1
		public static bool statReq2;	// true if Status of calculation is requested, Level 2
		public static bool statReq3;	// true if Status of calculation is requested, Level 3
		public static bool timeIt;		// timing mode turned on
		public static double time0;		// time in seconds of start of run
		public static double timeP;		// previous value of system clock for time recycle test
		public static double tV1;		// time value for delta time, previous value of tV2
		public static double tV2;		// time value for total time, delta = tV2 - tV1
		public static RichTextBox outS; // screen Output destination
		public static StreamWriter outL;// log file output #
		public static string nL;		// one Environment.NewLine \n or \r or \r\n
		public static string nL2;		// nL + nL
		public static string nLc;		// line continuation mark for WriteT nL or "-" + nL
	}
}
